@@ -18,8 +18,6 @@
 // SPDX-License-Identifier: LGPL-2.1-or-later
 
 use futures::channel::oneshot;
-use futures::future::BoxFuture;
-use futures::prelude::*;
 
 use gst::glib;
 use gst::prelude::*;
@@ -82,82 +80,80 @@ struct QueuePadSinkHandler;
 impl PadSinkHandler for QueuePadSinkHandler {
     type ElementImpl = Queue;
 
-    fn sink_chain(
+    async fn sink_chain(
         self,
         pad: gst::Pad,
         elem: super::Queue,
         buffer: gst::Buffer,
-    ) -> BoxFuture<'static, Result<gst::FlowSuccess, gst::FlowError>> {
-        async move {
-            gst::log!(CAT, obj = pad, "Handling {:?}", buffer);
-            let imp = elem.imp();
-            imp.enqueue_item(DataQueueItem::Buffer(buffer)).await
-        }
-        .boxed()
+    ) -> Result<gst::FlowSuccess, gst::FlowError> {
+        gst::log!(CAT, obj = pad, "Handling {:?}", buffer);
+        let imp = elem.imp();
+        imp.enqueue_item(DataQueueItem::Buffer(buffer)).await
     }
 
-    fn sink_chain_list(
+    async fn sink_chain_list(
         self,
         pad: gst::Pad,
         elem: super::Queue,
         list: gst::BufferList,
-    ) -> BoxFuture<'static, Result<gst::FlowSuccess, gst::FlowError>> {
-        async move {
-            gst::log!(CAT, obj = pad, "Handling {:?}", list);
-            let imp = elem.imp();
-            imp.enqueue_item(DataQueueItem::BufferList(list)).await
-        }
-        .boxed()
+    ) -> Result<gst::FlowSuccess, gst::FlowError> {
+        gst::log!(CAT, obj = pad, "Handling {:?}", list);
+        let imp = elem.imp();
+        imp.enqueue_item(DataQueueItem::BufferList(list)).await
     }
 
     fn sink_event(self, pad: &gst::Pad, imp: &Queue, event: gst::Event) -> bool {
-        gst::debug!(CAT, obj = pad, "Handling non-serialized {:?}", event);
+        gst::debug!(CAT, obj = pad, "Handling non-serialized {event:?}");
 
         if let gst::EventView::FlushStart(..) = event.view() {
-            if let Err(err) = imp.task.flush_start().await_maybe_on_context() {
-                gst::error!(CAT, obj = pad, "FlushStart failed {:?}", err);
-                gst::element_imp_error!(
-                    imp,
-                    gst::StreamError::Failed,
-                    ("Internal data stream error"),
-                    ["FlushStart failed {:?}", err]
-                );
+            if imp
+                .task
+                .flush_start()
+                .block_on_or_add_subtask_then(imp.obj(), |elem, res| {
+                    if let Err(err) = res {
+                        gst::error!(CAT, obj = elem, "FlushStart failed {err:?}");
+                        gst::element_error!(
+                            elem,
+                            gst::StreamError::Failed,
+                            ("Internal data stream error"),
+                            ["FlushStart failed {err:?}"]
+                        );
+                    }
+                })
+                .is_err()
+            {
                 return false;
             }
         }
 
-        gst::log!(CAT, obj = pad, "Forwarding non-serialized {:?}", event);
+        gst::log!(CAT, obj = pad, "Forwarding non-serialized {event:?}");
         imp.src_pad.gst_pad().push_event(event)
     }
 
-    fn sink_event_serialized(
+    async fn sink_event_serialized(
         self,
         pad: gst::Pad,
         elem: super::Queue,
         event: gst::Event,
-    ) -> BoxFuture<'static, bool> {
-        async move {
-            gst::log!(CAT, obj = pad, "Handling serialized {:?}", event);
+    ) -> bool {
+        gst::log!(CAT, obj = pad, "Handling serialized {:?}", event);
 
-            let imp = elem.imp();
+        let imp = elem.imp();
 
-            if let gst::EventView::FlushStop(..) = event.view() {
-                if let Err(err) = imp.task.flush_stop().await_maybe_on_context() {
-                    gst::error!(CAT, obj = pad, "FlushStop failed {:?}", err);
-                    gst::element_imp_error!(
-                        imp,
-                        gst::StreamError::Failed,
-                        ("Internal data stream error"),
-                        ["FlushStop failed {:?}", err]
-                    );
-                    return false;
-                }
+        if let gst::EventView::FlushStop(..) = event.view() {
+            if let Err(err) = imp.task.flush_stop().await {
+                gst::error!(CAT, obj = pad, "FlushStop failed {:?}", err);
+                gst::element_imp_error!(
+                    imp,
+                    gst::StreamError::Failed,
+                    ("Internal data stream error"),
+                    ["FlushStop failed {:?}", err]
+                );
             }
-
-            gst::log!(CAT, obj = pad, "Queuing serialized {:?}", event);
-            imp.enqueue_item(DataQueueItem::Event(event)).await.is_ok()
         }
-        .boxed()
+
+        gst::log!(CAT, obj = pad, "Queuing serialized {:?}", event);
+        imp.enqueue_item(DataQueueItem::Event(event)).await.is_ok()
     }
 
     fn sink_query(self, pad: &gst::Pad, imp: &Queue, query: &mut gst::QueryRef) -> bool {
@@ -181,31 +177,40 @@ impl PadSrcHandler for QueuePadSrcHandler {
     type ElementImpl = Queue;
 
     fn src_event(self, pad: &gst::Pad, imp: &Queue, event: gst::Event) -> bool {
-        gst::log!(CAT, obj = pad, "Handling {:?}", event);
+        gst::log!(CAT, obj = pad, "Handling {event:?}");
 
         use gst::EventView;
         match event.view() {
             EventView::FlushStart(..) => {
-                if let Err(err) = imp.task.flush_start().await_maybe_on_context() {
-                    gst::error!(CAT, obj = pad, "FlushStart failed {:?}", err);
-                }
+                let _ =
+                    imp.task
+                        .flush_start()
+                        .block_on_or_add_subtask_then(imp.obj(), |elem, res| {
+                            if let Err(err) = res {
+                                gst::error!(CAT, obj = elem, "FlushStart failed {err:?}");
+                            }
+                        });
             }
             EventView::FlushStop(..) => {
-                if let Err(err) = imp.task.flush_stop().await_maybe_on_context() {
-                    gst::error!(CAT, obj = pad, "FlushStop failed {:?}", err);
-                    gst::element_imp_error!(
-                        imp,
-                        gst::StreamError::Failed,
-                        ("Internal data stream error"),
-                        ["FlushStop failed {:?}", err]
-                    );
-                    return false;
-                }
+                let _ =
+                    imp.task
+                        .flush_stop()
+                        .block_on_or_add_subtask_then(imp.obj(), |elem, res| {
+                            if let Err(err) = res {
+                                gst::error!(CAT, obj = elem, "FlushStop failed {err:?}");
+                                gst::element_error!(
+                                    elem,
+                                    gst::StreamError::Failed,
+                                    ("Internal data stream error"),
+                                    ["FlushStop failed {err:?}"]
+                                );
+                            }
+                        });
             }
             _ => (),
         }
 
-        gst::log!(CAT, obj = pad, "Forwarding {:?}", event);
+        gst::log!(CAT, obj = pad, "Forwarding {event:?}");
         imp.sink_pad.gst_pad().push_event(event)
     }
 
@@ -224,12 +229,9 @@ impl PadSrcHandler for QueuePadSrcHandler {
             let (flags, min, max, align) = new_query.result();
             q.set(flags, min, max, align);
             q.add_scheduling_modes(
-                &new_query
+                new_query
                     .scheduling_modes()
-                    .iter()
-                    .cloned()
-                    .filter(|m| m != &gst::PadMode::Pull)
-                    .collect::<Vec<_>>(),
+                    .filter(|m| m != &gst::PadMode::Pull),
             );
             gst::log!(CAT, obj = pad, "Returning {:?}", q.query_mut());
             return true;
@@ -279,109 +281,98 @@ impl QueueTask {
 impl TaskImpl for QueueTask {
     type Item = DataQueueItem;
 
-    fn start(&mut self) -> BoxFuture<'_, Result<(), gst::ErrorMessage>> {
-        async move {
-            gst::log!(CAT, obj = self.element, "Starting task");
-
-            let queue = self.element.imp();
-            let mut last_res = queue.last_res.lock().unwrap();
-
-            self.dataqueue.start();
-
-            *last_res = Ok(gst::FlowSuccess::Ok);
-
-            gst::log!(CAT, obj = self.element, "Task started");
-            Ok(())
-        }
-        .boxed()
+    fn obj(&self) -> &impl IsA<glib::Object> {
+        &self.element
     }
 
-    fn try_next(&mut self) -> BoxFuture<'_, Result<DataQueueItem, gst::FlowError>> {
-        async move {
-            self.dataqueue
-                .next()
-                .await
-                .ok_or_else(|| panic!("DataQueue stopped while Task is Started"))
+    async fn start(&mut self) -> Result<(), gst::ErrorMessage> {
+        gst::log!(CAT, obj = self.element, "Starting task");
+
+        let queue = self.element.imp();
+        let mut last_res = queue.last_res.lock().unwrap();
+
+        if let Some(pending_queue) = queue.pending_queue.lock().unwrap().as_mut() {
+            pending_queue.notify_more_queue_space();
         }
-        .boxed()
+
+        self.dataqueue.start();
+
+        *last_res = Ok(gst::FlowSuccess::Ok);
+
+        gst::log!(CAT, obj = self.element, "Task started");
+        Ok(())
     }
 
-    fn handle_item(&mut self, item: DataQueueItem) -> BoxFuture<'_, Result<(), gst::FlowError>> {
-        async {
-            let res = self.push_item(item).await;
-            let queue = self.element.imp();
-            match res {
-                Ok(()) => {
-                    gst::log!(CAT, obj = self.element, "Successfully pushed item");
-                    *queue.last_res.lock().unwrap() = Ok(gst::FlowSuccess::Ok);
-                }
-                Err(gst::FlowError::Flushing) => {
-                    gst::debug!(CAT, obj = self.element, "Flushing");
-                    *queue.last_res.lock().unwrap() = Err(gst::FlowError::Flushing);
-                }
-                Err(gst::FlowError::Eos) => {
-                    gst::debug!(CAT, obj = self.element, "EOS");
-                    *queue.last_res.lock().unwrap() = Err(gst::FlowError::Eos);
-                    queue.src_pad.push_event(gst::event::Eos::new()).await;
-                }
-                Err(err) => {
-                    gst::error!(CAT, obj = self.element, "Got error {}", err);
-                    gst::element_error!(
-                        &self.element,
-                        gst::StreamError::Failed,
-                        ("Internal data stream error"),
-                        ["streaming stopped, reason {}", err]
-                    );
-                    *queue.last_res.lock().unwrap() = Err(err);
-                }
+    async fn try_next(&mut self) -> Result<DataQueueItem, gst::FlowError> {
+        self.dataqueue
+            .next()
+            .await
+            .ok_or_else(|| panic!("DataQueue stopped while Task is Started"))
+    }
+
+    async fn handle_item(&mut self, item: DataQueueItem) -> Result<(), gst::FlowError> {
+        let res = self.push_item(item).await;
+        let queue = self.element.imp();
+        match res {
+            Ok(()) => {
+                gst::log!(CAT, obj = self.element, "Successfully pushed item");
+                *queue.last_res.lock().unwrap() = Ok(gst::FlowSuccess::Ok);
+            }
+            Err(gst::FlowError::Flushing) => {
+                gst::debug!(CAT, obj = self.element, "Flushing");
+                *queue.last_res.lock().unwrap() = Err(gst::FlowError::Flushing);
+            }
+            Err(gst::FlowError::Eos) => {
+                gst::debug!(CAT, obj = self.element, "EOS");
+                *queue.last_res.lock().unwrap() = Err(gst::FlowError::Eos);
+                queue.src_pad.push_event(gst::event::Eos::new()).await;
+            }
+            Err(err) => {
+                gst::error!(CAT, obj = self.element, "Got error {}", err);
+                gst::element_error!(
+                    &self.element,
+                    gst::StreamError::Failed,
+                    ("Internal data stream error"),
+                    ["streaming stopped, reason {}", err]
+                );
+                *queue.last_res.lock().unwrap() = Err(err);
             }
-
-            res
         }
-        .boxed()
+
+        res
     }
 
-    fn stop(&mut self) -> BoxFuture<'_, Result<(), gst::ErrorMessage>> {
-        async move {
-            gst::log!(CAT, obj = self.element, "Stopping task");
-
-            let queue = self.element.imp();
-            let mut last_res = queue.last_res.lock().unwrap();
-
-            self.dataqueue.stop();
-            self.dataqueue.clear();
-
-            if let Some(mut pending_queue) = queue.pending_queue.lock().unwrap().take() {
-                pending_queue.notify_more_queue_space();
-            }
-
-            *last_res = Err(gst::FlowError::Flushing);
-
-            gst::log!(CAT, obj = self.element, "Task stopped");
-            Ok(())
-        }
-        .boxed()
+    async fn stop(&mut self) -> Result<(), gst::ErrorMessage> {
+        gst::log!(CAT, obj = self.element, "Stopping task");
+        self.flush_start().await?;
+        gst::log!(CAT, obj = self.element, "Task stopped");
+        Ok(())
     }
 
-    fn flush_start(&mut self) -> BoxFuture<'_, Result<(), gst::ErrorMessage>> {
-        async move {
-            gst::log!(CAT, obj = self.element, "Starting task flush");
+    async fn flush_start(&mut self) -> Result<(), gst::ErrorMessage> {
+        gst::log!(CAT, obj = self.element, "Task flush start");
 
-            let queue = self.element.imp();
-            let mut last_res = queue.last_res.lock().unwrap();
+        let queue = self.element.imp();
+        let mut last_res = queue.last_res.lock().unwrap();
 
-            self.dataqueue.clear();
+        self.dataqueue.stop();
+        self.dataqueue.clear();
 
-            if let Some(mut pending_queue) = queue.pending_queue.lock().unwrap().take() {
-                pending_queue.notify_more_queue_space();
-            }
+        *last_res = Err(gst::FlowError::Flushing);
 
-            *last_res = Err(gst::FlowError::Flushing);
-
-            gst::log!(CAT, obj = self.element, "Task flush started");
-            Ok(())
+        if let Some(mut pending_queue) = queue.pending_queue.lock().unwrap().take() {
+            pending_queue.notify_more_queue_space();
         }
-        .boxed()
+
+        gst::log!(CAT, obj = self.element, "Task flush started");
+        Ok(())
+    }
+
+    async fn flush_stop(&mut self) -> Result<(), gst::ErrorMessage> {
+        gst::log!(CAT, obj = self.element, "Task flush stop");
+        self.start().await?;
+        gst::log!(CAT, obj = self.element, "Task flush stopped");
+        Ok(())
     }
 }
 
@@ -502,6 +493,13 @@ impl Queue {
                     .unwrap_or(true)
                 {
                     if pending_queue.is_none() {
+                        // Lock order: last_res, then pending_queue
+                        drop(pending_queue);
+                        if *self.last_res.lock().unwrap() == Err(gst::FlowError::Flushing) {
+                            return Err(gst::FlowError::Flushing);
+                        }
+                        pending_queue = self.pending_queue.lock().unwrap();
+
                         *pending_queue = Some(PendingQueue {
                             more_queue_space_sender: None,
                             scheduled: false,
@@ -574,8 +572,6 @@ impl Queue {
             },
         );
 
-        *self.dataqueue.lock().unwrap() = Some(dataqueue.clone());
-
         let context =
             Context::acquire(&settings.context, settings.context_wait).map_err(|err| {
                 gst::error_msg!(
@@ -584,40 +580,53 @@ impl Queue {
                 )
             })?;
 
+        *self.dataqueue.lock().unwrap() = Some(dataqueue.clone());
+
         self.task
             .prepare(QueueTask::new(self.obj().clone(), dataqueue), context)
-            .block_on()?;
-
-        gst::debug!(CAT, imp = self, "Prepared");
-
-        Ok(())
+            .block_on_or_add_subtask_then(self.obj(), |elem, res| {
+                if res.is_ok() {
+                    gst::debug!(CAT, obj = elem, "Prepared");
+                }
+            })
     }
 
     fn unprepare(&self) {
         gst::debug!(CAT, imp = self, "Unpreparing");
+        let _ = self
+            .task
+            .unprepare()
+            .block_on_or_add_subtask_then(self.obj(), |elem, _| {
+                let imp = elem.imp();
+                *imp.dataqueue.lock().unwrap() = None;
+                *imp.pending_queue.lock().unwrap() = None;
 
-        self.task.unprepare().block_on().unwrap();
+                *imp.last_res.lock().unwrap() = Ok(gst::FlowSuccess::Ok);
 
-        *self.dataqueue.lock().unwrap() = None;
-        *self.pending_queue.lock().unwrap() = None;
-
-        *self.last_res.lock().unwrap() = Ok(gst::FlowSuccess::Ok);
-
-        gst::debug!(CAT, imp = self, "Unprepared");
+                gst::debug!(CAT, obj = elem, "Unprepared");
+            });
     }
 
     fn stop(&self) -> Result<(), gst::ErrorMessage> {
         gst::debug!(CAT, imp = self, "Stopping");
-        self.task.stop().await_maybe_on_context()?;
-        gst::debug!(CAT, imp = self, "Stopped");
-        Ok(())
+        self.task
+            .stop()
+            .block_on_or_add_subtask_then(self.obj(), |elem, res| {
+                if res.is_ok() {
+                    gst::debug!(CAT, obj = elem, "Stopped");
+                }
+            })
     }
 
     fn start(&self) -> Result<(), gst::ErrorMessage> {
         gst::debug!(CAT, imp = self, "Starting");
-        self.task.start().await_maybe_on_context()?;
-        gst::debug!(CAT, imp = self, "Started");
-        Ok(())
+        self.task
+            .start()
+            .block_on_or_add_subtask_then(self.obj(), |elem, res| {
+                if res.is_ok() {
+                    gst::debug!(CAT, obj = elem, "Started");
+                }
+            })
     }
 }
 
@@ -677,6 +686,21 @@ impl ObjectImpl for Queue {
                     .maximum(u64::MAX - 1)
                     .default_value(DEFAULT_MAX_SIZE_TIME.nseconds())
                     .build(),
+                glib::ParamSpecUInt::builder("current-level-buffers")
+                    .nick("Current Level Buffers")
+                    .blurb("Current number of buffers in the queue")
+                    .read_only()
+                    .build(),
+                glib::ParamSpecUInt::builder("current-level-bytes")
+                    .nick("Current Level Bytes")
+                    .blurb("Current amount of data in the queue (bytes)")
+                    .read_only()
+                    .build(),
+                glib::ParamSpecUInt64::builder("current-level-time")
+                    .nick("Current Level Time")
+                    .blurb("Current amount of data in the queue (in ns)")
+                    .read_only()
+                    .build(),
             ]
         });
 
@@ -711,13 +735,42 @@ impl ObjectImpl for Queue {
     }
 
     fn property(&self, _id: usize, pspec: &glib::ParamSpec) -> glib::Value {
-        let settings = self.settings.lock().unwrap();
         match pspec.name() {
-            "max-size-buffers" => settings.max_size_buffers.to_value(),
-            "max-size-bytes" => settings.max_size_bytes.to_value(),
-            "max-size-time" => settings.max_size_time.nseconds().to_value(),
-            "context" => settings.context.to_value(),
-            "context-wait" => (settings.context_wait.as_millis() as u32).to_value(),
+            "max-size-buffers" => self.settings.lock().unwrap().max_size_buffers.to_value(),
+            "max-size-bytes" => self.settings.lock().unwrap().max_size_bytes.to_value(),
+            "max-size-time" => self
+                .settings
+                .lock()
+                .unwrap()
+                .max_size_time
+                .nseconds()
+                .to_value(),
+            "current-level-buffers" => self
+                .dataqueue
+                .lock()
+                .unwrap()
+                .as_ref()
+                .map_or(0, |d| d.cur_level_buffers())
+                .to_value(),
+            "current-level-bytes" => self
+                .dataqueue
+                .lock()
+                .unwrap()
+                .as_ref()
+                .map_or(0, |d| d.cur_level_bytes())
+                .to_value(),
+            "current-level-time" => self
+                .dataqueue
+                .lock()
+                .unwrap()
+                .as_ref()
+                .map_or(gst::ClockTime::ZERO, |d| d.cur_level_time())
+                .nseconds()
+                .to_value(),
+            "context" => self.settings.lock().unwrap().context.to_value(),
+            "context-wait" => {
+                (self.settings.lock().unwrap().context_wait.as_millis() as u32).to_value()
+            }
             _ => unimplemented!(),
         }
     }
